package urpmchecker;

use strict;

my $VERSION = "0.1.0";
my $CVS = '$Id: urpmchecker.pm,v 1.51 2003/08/28 22:16:37 othauvin Exp $';

use URPM;
use URPM::Build;
use MDK::Common;
use BerkeleyDB;

=begin

=head1 NAME

urpmchecker - function to works on rpms repository

=head1 SYMOPSYS

use urpmchecker;
$urpmchecker = new urpmchecker;

	
=head1 SEE ALSO

perl-URPM

=head1 LICENSE

Olivier Thauvin <thauvin@aerov.jussieu.fr>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

=head1 DESCRIPTION

C<urpmchecker> include urpmchecker functions

=cut

=item * $urpmchecker->new($param);

    $param = {
              rpmrc => '/usr/lib/rpm/rpmrc', #rpmrc to use
              arch => 'arch', # set arch
    }

=cut

sub new {
	my ($self, $param) = @_;
	my $env = BerkeleyDB::Env->new('Home' => $ENV{TMPDIR} || $ENV{TMP} || "/tmp");
	$param->{dbfiles} =  BerkeleyDB::Btree->new(
                    'Flags'    => DB_CREATE,
                    'Property' => DB_DUP | DB_DUPSORT,
                    'Env' => $env)
        		or die "Cannot open db: $! $BerkeleyDB::Error\n";
	$param->{rpmvalue} = parse_rc($param->{rpmrc} || "/usr/lib/rpm/rpmrc");
	
	bless {
		urpm => URPM->new(),
		dbfiles => $param->{dbfiles},
		dbfilesf => $param->{dbfilesf},
        dbprovides => {},
		dbprovidesf => $param->{dbprovidesf},
        dbobsoletes => {},
        # ARRAY array{idpackage} = [ pkglist ]
        obsoletes => {},
		deps => {},
        conflicts => {},
        #
		repository => [],
		id_level => {},
		id_rep => {},
		error => {},
		rpmvalues => $param->{rpmvalue},
		'arch' => $param->{arch}
	}, $self;
}

sub DESTROY {
   my ($self) = @_;
   $self->{dbfiles}->db_close();
   unlink($self->{dbfilesf}); 
}

=item * $urpmchecker::version();

return module version.

=cut 

sub version {
	"$VERSION"
}

=item * $urpmchecker->add($item, $param);

$item is a rpm, a dir, or a hdlist

    $param = {
              nextlevel => 1,   # level ++
              check => 1,       # check this repository
              chksrc => 1,      # check src <=> bin
              nochkdep => 1,    # don't check dependencies
              nochkcf => 1,     # don't check files conflict
              chkreb => 1, # check package no up2date with src, need to be build
             }

=cut

sub add {
	my ($self, $item, $to_add) = @_;
    my $last_rep_level = $self->last_rep_level;
	$to_add->{nextlevel} || $last_rep_level==0 and $last_rep_level++;
	print STDERR "Add $item in list\n" if $self->{verbose};
	$to_add->{level} =  $last_rep_level;
	$to_add->{origin} = $item;
			
	if ($item =~ /cz$/) { 
         -f $item or return(undef);
         $to_add->{hdlist} = $item;
     } elsif (-f $item && $item =~ /\.rpm$/) {
         @{$to_add->{rpms}} = $item;
     } else {
         my @l = glob "$item/*.rpm";
         scalar(@l) or return(undef);
         @{$to_add->{rpms}} = @l;
     }
     push @{$self->{repository}}, { %$to_add };
     1
}

sub last_rep_level {
    my ($self) = @_;
    my $level=0;
    foreach (@{$self->{repository}}) {
        $level = $_->{level} if ($level < $_->{level});
    }
    $level
}

=item * $urpmchecker->parse_list();

For each repository added, parse rpm headers and store information internally.

=cut

sub parse_list {
	my ($self) = @_;
	foreach (@{ $self->{repository} }) {
         if ($_->{hdlist}) {
             ($_->{start}, $_->{end}) = $self->{urpm}->parse_hdlist($_->{hdlist});
             push @{$_->{list}}, $_->{start} .. $_->{end};
         }
         @{$_->{list}} = map { $self->{urpm}->parse_rpm($_, keep_all_tags => 1) } @{$_->{rpms}} if $_->{rpms};
	}
	foreach my $n (0 .. $#{$self->{repository}}) {
		my $l = $self->{repository}[$n];
		@{$l->{list}} = uniq(@{$l->{list}});
		#removing non compatible rpm
		my @temp = grep {
              my $pkg = $self->{urpm}{depslist}[$_];
              $pkg->arch ne 'src' ?
                $self->rpmcompat($pkg->arch, "arch_compat") : # FIXME: missing $type and $os_arch parameters
                  $self->rpmisbuildable($pkg);

          } @{$l->{list}};
		@{$l->{list}} = @temp;
		$self->{id_level}{$_} = $l->{level} foreach @{$l->{list}};
		$self->{id_rep}{$_} = $n foreach @{ $l->{list} };
	}
	foreach my $l (@{$self->{repository}}) {
		foreach my $u_rpm (@{$l->{list}}) {
			my $pkg = $self->{urpm}{depslist}[$u_rpm];
			my ($name, $version, $release, $arch);
			if ($pkg->arch eq 'src') {
				($name, $version, $release) = ($pkg->name, $pkg->version, $pkg->release);
				$arch = 'src';
				push @{$self->{rpm}{source}{$pkg->name}}, $u_rpm;
			} else {
				my $sourcerpm = $pkg->sourcerpm;
				($name, $version, $release) = $sourcerpm =~ /^(.*)-(.*)-(.*)\.src\.rpm$/; # no check for faillure!!!
				$arch = 'bin';
				push @{$self->{rpm}{binary}{$pkg->name}}, $u_rpm;
			}
			push @{$self->{rpm}{$arch}{$name}{"$version-$release"}}, $u_rpm;
		}
	}
	
	foreach my $l (@{$self->{repository}}) {
		foreach my $u_rpm (@{$l->{list}}) {
			my $pkg = $self->{urpm}{depslist}[$u_rpm];
			# All info after does concern src... why wast time
			# is a src can obsoletes another ??? That is the question. It seems no
			$pkg->arch eq 'src' and next;
			my $best = $u_rpm;
            # because rpm < 4, packages don't provides theirself
            push @{$self->{dbprovides}{$pkg->name}}, "$u_rpm == " . ($pkg->epoch ? $pkg->epoch.':' : '') . $pkg->version."-".$pkg->release;
			my @p = $pkg->provides;
			foreach (@p) {
				my ($n, $o, $v) = /^([^\s\[]*)(?:\[\*\])?\s*\[?([^\s\]]*)\s*([^\s\]]*)/;
                push @{$self->{dbprovides}{$n}}, "$u_rpm $o $v";
			}

			foreach (@{$self->{rpm}{binary}{$pkg->name}}) {
				my $pkg2 = $self->{urpm}{depslist}[$_];
                $self->rpmcompat($pkg2->arch, "arch_compat", $pkg->arch) or next;
				$pkg->compare_pkg($pkg2) < 0 and $best = $_;
			}
			$best == $u_rpm or next;
			
            my @o = $pkg->obsoletes;
            foreach (@o) {
            	my ($n, $o, $v) = /^([^\s\[]*)(?:\[\*\])?\s*\[?([^\s\]]*)\s*([^\s\]]*)/;
            	push @{$self->{dbobsoletes}{$n}}, "$u_rpm $o $v";
            }
			
            my @f = $pkg->files;
			my @m = $pkg->files_mode;
			my @u = $pkg->files_owner;
			my @g = $pkg->files_group;
			my @md5 = $pkg->files_md5sum;
			foreach (0 .. $#f) {
				$self->{dbfiles}->db_put($f[$_], "$u_rpm $m[$_] $md5[$_] $u[$_] $g[$_]");
			}
		}
	}
	foreach my $l (@{$self->{repository}}) {
		foreach my $u_rpm (@{$l->{list}}) {
			my $pkg = $self->{urpm}{depslist}[$u_rpm];
			foreach ($pkg->requires) {
				my ($n, $o, $v) = /^([^\s\[]*)(?:\[\*\])?\s*\[?([^\s\]]*)\s*([^\s\]]*)/;
				$o =~ s/==/=/g;
				my ($value, $cursor_state, @tdeps) = (undef, DB_SET, ());
				if ($n =~ m!^/!) {
					my $cursor = $self->{dbfiles}->db_cursor();
					while ($cursor->c_get($n, $value, $cursor_state) == 0) {
						push @tdeps, (split(' ', $value))[0];
						$cursor_state = DB_NEXT_DUP;
					}
                } else {
                    foreach $value (@{$self->{dbprovides}{$n}}) {
                        my ($p, $po, $pv) = split(' ', $value);
                        $po =~ s/==/=/g;
                        URPM::ranges_overlap("a $po $pv", "a $o $v") &&
                            ($self->rpmcompat($pkg->arch, "arch_compat", $self->{urpm}{depslist}[$p]->arch) ||
                                $self->rpmcompat($self->{urpm}{depslist}[$p]->arch, "arch_compat", $pkg->arch)) and push @tdeps, $p;
                    }
				}
                @tdeps = uniq(@tdeps);
                push @{$self->{deps}{$u_rpm}}, (scalar @tdeps) ? join('|', @tdeps) : "NOT$n $o $v";
			}
			
            # Parsing obsoletes
            foreach (@{$self->{dbobsoletes}{$pkg->name}}) {
            	my ($p, $o, $v) = split(' ');
            	$o && ! eval($pkg->compare($v) . $o . 0) and next;
            	$pkg->arch eq 'src' || $self->{urpm}{depslist}[$p]->arch eq 'src' and
                      $pkg->arch ne $self->{urpm}{depslist}[$p]->arch and next;
            	push @{$self->{obsoletes}{$u_rpm}}, $p;
            }
            @{$self->{obsoletes}{$u_rpm}} = uniq(@{$self->{obsoletes}{$u_rpm}});

            # No, obsoletes are NOT expand to provides
            # Obsoletes
#			foreach ($pkg->obsoletes) {
#                #print $pkg->fullname ."$_\n";
#				my ($n, $o, $v) = /^([^\s\[]*)(?:\[\*\])?\s*\[?([^\s\]]*)\s*([^\s\]]*)/;
#				$o =~ s/==/=/g;
#				my ($value, $cursor_state) = (undef, DB_SET);
#				my $cursor = $self->{dbprovides}->db_cursor();
#				while ($cursor->c_get($n, $value, $cursor_state) == 0) {
#					my ($p, $po, $pv) = split(' ', $value);
#					$po =~ s/==/=/g;
#					# It want a name, well, we not loose time with long name, we allready make it ;)
#					$u_rpm != $p && URPM::ranges_overlap("a $po $pv", "a $o $v") && 
#                        ($self->rpmcompat($pkg->arch, "arch_compat", $self->{urpm}{depslist}[$p]->arch) || $self->rpmcompat($self->{urpm}{depslist}[$p]->arch, "arch_compat", $pkg->arch)) and do {
#                            push @{$self->{obsoletes}{$p}}, $u_rpm;
#                            @{$self->{obsoletes}{$p}} = uniq(@{$self->{obsoletes}{$p}});
#                    };
#					$cursor_state = DB_NEXT_DUP;
#				}
#			}
#            @{$self->{obsoletes}{$_}} = uniq(@{$self->{obsoletes}{$_}}) foreach(keys %{$self->{obsoletes}});
            
            
            # Conflict:
            foreach ($pkg->conflicts) {
            	my ($n, $o, $v) = /^([^\s\[]*)(?:\[\*\])?\s*\[?([^\s\]]*)\s*([^\s\]]*)/;
            	$o =~ s/==/=/g;
            	my ($value, $cursor_state) = (undef, DB_SET);
            	foreach $value (@{$self->{dbprovides}{$n}}) {
                    my ($p, $po, $pv) = split(' ', $value);
            		$po =~ s/==/=/g;
            		# It want a name, well, we not loose time with long name, we allready make it ;)
            		URPM::ranges_overlap("a $po $pv", "a $o $v") and do {
                       $self->rpmcompat($pkg->arch, "arch_compat", $self->{urpm}{depslist}[$p]->arch) and push(@{$self->{conflicts}{$p}}, $u_rpm); 
                       $self->rpmcompat($self->{urpm}{depslist}[$p]->arch, "arch_compat", $pkg->arch) and push(@{$self->{conflicts}{$u_rpm}}, $p);
                   };
            	}
            }

            @{$self->{conflicts}{$_}} = uniq(@{$self->{conflicts}{$_}}) foreach(keys %{$self->{conflicts}});
            
		}
	}
}
sub analyze {
	my ($self) = @_;
	$self->analyze_dep();
	$self->analyze_files();
	$self->analyze_src();
}

sub analyze_dep {
	my ($self) = @_;
	foreach my $l (@{$self->{repository}}) {
		$l->{check} or next;
		$l->{nochkdep} and next;
		foreach my $u_rpm (@{$l->{list}}) {
			my $pkg = $self->{urpm}{depslist}[$u_rpm];
			my $cat = $pkg->arch eq 'src' ? 'source' : 'binary';

			foreach (@{$self->{rpm}{$cat}{$pkg->name}}) {
				$u_rpm == $_ and next;
				my $pkg2 = $self->{urpm}{depslist}[$_];
                $pkg->arch ne $pkg2->arch && $pkg->arch ne 'noarch' && $pkg2->arch ne 'noarch' and next;
                my $cmp = $pkg2->compare_pkg($pkg);
                if ($cmp == 0) {
                    $self->{id_rep}{$u_rpm} < $self->{id_rep}{$_} and next;
                    $self->report_error($u_rpm, { id => $_, type => 'VERT', msg => "allready find" });
                    next;
                } elsif ($cmp > 0) {
					$self->idupdate($_) and next;
					$self->report_error($u_rpm, { id => $_, type => 'VER', msg => "new version" });
                         next if $_ ne $u_rpm;
				};
			}
			
			if ($pkg->arch ne 'src') {
			    my ($name, $version, $release) = $pkg->sourcerpm =~ /^(.*)-(.*)-(.*)\.src\.rpm$/;

			    foreach my $vr (keys %{$self->{rpm}{bin}{$name}}) {
				    $vr eq "$version-$release" and next;
				    foreach (@{$self->{rpm}{bin}{$name}{$vr}}) {
					    $self->idupdate($_) and next;
					    my $pkg2 = $self->{urpm}{depslist}[$_];
					    $pkg->name eq $pkg2->name and next;
					    URPM::ranges_overlap("a == $version-$release", "a < $vr") && ($self->rpmcompat($pkg->arch, "arch_compat", $pkg2->arch) || $self->rpmcompat($pkg2->arch, "arch_compat", $pkg->arch)) and $self->report_error($u_rpm, { msgsum => "$name-$vr.src", type => 'VERM', msg => "seems to be old " });

				    }
			    }
			}

			foreach my $dependancy (@{$self->{deps}{$u_rpm}}) {
                $dependancy =~ /^NOT/ and do {
                    my ($dep) = $dependancy =~ m/^NOT(.*)/;
                    $self->report_error($u_rpm, { msgsum => $dep, type => 'DEP' }); 
                    next;
                };
                my $best = $self->find_best_pkg(split(/\|/,$dependancy));
                $self->is_best_pkg($best) and $self->report_error($u_rpm, { id => $best, type => 'DEPB', msg => "dep on Obs packages" });
				$self->idupdate($best) and next;
				$self->report_error($u_rpm, { id => $best, type => 'DEPO', msg => "req in other level" }) if $l->{level} < $self->idlevel($best);
			}
			foreach (@{$self->{obsoletes}{$u_rpm}}) {
				$self->report_error($u_rpm, { id => $_, type => 'OBS', msg => "obs by" }) if $l->{level} >= $self->{id_level}{$_};
			}			

            # checking if install possible
            #
            #foreach my $need ($self->pkg_need({ rpm => $u_rpm })) {
            #    any { $need == $_ } @{$self->{conflicts}{$u_rpm}} and $self->report_error($u_rpm, { id => $need, type => 'DEPC', msg => "need but conflict" });
            #}
		}
	}
}	

sub analyze_files {
	my ($self) = @_;	
    my (%conflicting_cache,%pkg_need_cache);
	
	my ($cursor, $file, $filekey) = $self->{dbfiles}->db_cursor();
	while ($cursor->c_get($file, $filekey, DB_NEXT) == 0) {
		my @l;
        my $fetch = sub {
              push @l, { u_rpm => $1, mode => $2, md5sum => $3, user => $4, group => $5 } if $filekey =~ m/(\S*) (\S*) (\S*) (\S*) (\S*)/;
        };
        $fetch->();
		$fetch->() while $cursor->c_get($file, $filekey, DB_NEXT_DUP) == 0;
		
		foreach my $file_entry (@l) {
			$self->{repository}[$self->{id_rep}{$file_entry->{u_rpm}}]{check} or next;
			$self->{repository}[$self->{id_rep}{$file_entry->{u_rpm}}]{nochkcf} and next;
		
			my $pkg = $self->{urpm}{depslist}[$file_entry->{u_rpm}];
			my $best = $file_entry->{u_rpm};
			foreach (@{$self->{rpm}{binary}{$pkg->name}}) {
				my $pkg2 = $self->{urpm}{depslist}[$_];
				$pkg->compare_pkg($pkg2) < 1 and $best = $_;
			}
			$best == $file_entry->{u_rpm} or next;

			# Is the parent is provide by another rpm
			my ($reper, $refile) = $file =~  m!^(.*)/(.*)!;
            while ($reper) {
			    my $cursor2 = $self->{dbfiles}->db_cursor();
			
			    my ($cursor_state, $value, @d) = (DB_SET);
			    while ($cursor2->c_get($reper, $value, $cursor_state) == 0) {
				    my ($p, $mode) = $value =~ m/(\S*) (\S*) (\S*) (\S*) (\S*)/;
				    ($mode & 0170000) == 0040000 and push @d, $p;
				    $cursor_state = DB_NEXT_DUP;	
			    }
			    if (@d) {
				    my $min = $self->find_best_pkg(@d);
				    $self->{id_level}{$min} > $self->{id_level}{$file_entry->{u_rpm}} and $self->report_error($file_entry->{u_rpm}, { type => 'DIRO', msg => "file $refile, rep in other level", msgsum => "$reper/" });
                    last;
			    } else {
				    $self->report_error($file_entry->{u_rpm}, { type => 'DIRM', msg => "file $refile", msgsum => "$reper/" });
			    }
                ($reper, $refile) = $reper =~  m!^(.*)/(.*)!;
			}
			
			foreach my $tmpfentry (@l) {
				# Check rpm against rpm
				$tmpfentry->{u_rpm} eq $file_entry->{u_rpm} and next; # Check itself, no !
				
                $self->{id_level}{$file_entry->{u_rpm}} < $self->{id_level}{$tmpfentry->{u_rpm}} and next;
				
				# same name, surelly an update
				$self->{urpm}{depslist}[$tmpfentry->{u_rpm}]->name eq $self->{urpm}{depslist}[$file_entry->{u_rpm}]->name and next;

                # package is obsoleted or obsolete or we are not comparing latest package
                $self->is_best_pkg($tmpfentry->{u_rpm}) || $self->is_best_pkg($file_entry->{u_rpm}) and next;
			
				# Is those packages conflicts ?
                #pkg_is_conflict($self->{urpm}{depslist}[$tmpfentry->{u_rpm}], $self->{urpm}{depslist}[$file_entry->{u_rpm}]) and next;
                if (!defined($conflicting_cache{$file_entry->{u_rpm}}{$tmpfentry->{u_rpm}})) {
                    my $cfl_yes=0;
                    # print $self->{urpm}{depslist}[$file_entry->{u_rpm}]->fullname."\n";
                    my @a = defined $pkg_need_cache{$file_entry->{u_rpm}} ? @{$pkg_need_cache{$file_entry->{u_rpm}}} : (@{$pkg_need_cache{$file_entry->{u_rpm}}} = $self->pkg_need({ rpm => $file_entry->{u_rpm} }));
                    my @b = defined $pkg_need_cache{$tmpfentry->{u_rpm}} ? @{$pkg_need_cache{$tmpfentry->{u_rpm}}} : (@{$pkg_need_cache{$tmpfentry->{u_rpm}}} = $self->pkg_need({ rpm => $tmpfentry->{u_rpm} }));
                    foreach my $aa (difference2(@a,@b),$file_entry->{u_rpm}) {
                        foreach my $bb (difference2(@b,@a),$tmpfentry->{u_rpm}) {
                            any { $aa == $_ } @{$self->{conflicts}{$bb}} and $cfl_yes = 1;
                            $cfl_yes and last;
                        }
                        $cfl_yes and last;
                    }
                    $conflicting_cache{$tmpfentry->{u_rpm}}{$file_entry->{u_rpm}} = $conflicting_cache{$file_entry->{u_rpm}}{$tmpfentry->{u_rpm}} = $cfl_yes;
                }
                $conflicting_cache{$file_entry->{u_rpm}}{$tmpfentry->{u_rpm}} == 1 and next;
                
                # Now checking files against files, here the tags:	
				#define __S_IFDIR       0040000 /* Directory.  */
				#define __S_IFCHR       0020000 /* Character device.  */
				#define __S_IFBLK       0060000 /* Block device.  */
				#define __S_IFREG       0100000 /* Regular file.  */
				#define __S_IFIFO       0010000 /* FIFO.  */
				#define __S_IFLNK       0120000 /* Symbolic link.  */
				#define __S_IFSOCK      0140000 /* Socket.  */
						
				($tmpfentry->{mode} & 0170000) != ($file_entry->{mode} & 0170000) and do { 
					$self->report_error($file_entry->{u_rpm}, { id => $tmpfentry->{u_rpm}, type => 'CFLT', msg => $file }); 
					next 
					};

				($tmpfentry->{mode} & 0006666) != ($file_entry->{mode} & 0006666) and
					$self->report_error($file_entry->{u_rpm}, { id => $tmpfentry->{u_rpm}, type => 'CFLP', msg => $file });
				($tmpfentry->{user} ne $file_entry->{user} || $tmpfentry->{group} ne $file_entry->{group}) and
					$self->report_error($file_entry->{u_rpm}, { id => $tmpfentry->{u_rpm}, type => 'CFLO', msg => $file });
				($file_entry->{mode} & 0170000) == 0040000 and next;
						
				$tmpfentry->{md5sum} eq $file_entry->{md5sum} and next;
								
				$self->report_error($file_entry->{u_rpm}, { id => $tmpfentry->{u_rpm}, type => 'CFL', msg => $file });
			}
		}
	}
	undef $cursor;
}

sub analyze_src {
	my ($self) = @_;
	foreach my $repository (@{$self->{repository}}) {
		$repository->{check} or next;
		$repository->{chksrc} || $repository->{chkreb} or next;
		foreach my $u_rpm (@{$repository->{list}}) {
			my $pkg = $self->{urpm}{depslist}[$u_rpm];
			if ($pkg->arch eq 'src') {
                if (exists $self->{rpm}{bin}{$pkg->name}{$pkg->version.'-'.$pkg->release}) {
                    any { $self->{id_level}{$_} == $repository->{level} and $_ } @{$self->{rpm}{bin}{$pkg->name}{$pkg->version.'-'.$pkg->release}} and next;
				    foreach (@{$self->{rpm}{bin}{$pkg->name}{$pkg->version.'-'.$pkg->release}}) {
						$self->report_error($u_rpm, { type => 'BINO', msg => sprintf("bin in [%s]", $self->{id_level}{$_}), id => $_ });
                    }
                } else {
                    if ($repository->{chkreb}) {
                        $self->report_error($u_rpm, { type => 'REB', msg => "Need to be build" });
                    } else {
				        exists $self->{rpm}{bin}{$pkg->name} or $self->report_error($u_rpm, { type => 'BIN', msg => "no binary from " . $pkg->name });
                    }
                } 
			} else {
				my ($name, $version, $release) = $pkg->sourcerpm =~ /^(.*)-(.*)-(.*)\.src\.rpm$/;
				$self->{rpm}{src}{$name} or $self->report_error($u_rpm, { type => 'SRC', msg => "no src named $name" });
                foreach (grep { $self->{id_level}{$_} != $repository->{level} && !$repository->{update} and $_ } @{$self->{rpm}{src}{$name}{"$version-$release"}}) {
					$self->report_error($u_rpm, { type => 'SRCO', msg => "src in [$self->{id_level}{$_}]", id => $_ });
				}
			}
		}
	}
	
}

=item * $urpmchecker->report_error($self, $id, $error);

    $id = id of rpm
    error = {
             id = id of rpm
             +fullname = name of rpm
             +packager => from latest changelog
             +sourcerpm => the src
             +level => repository level
             +sname => name tag of source package
             @err = 
             { 
              id => rpm wich conflict
              +fullname => of package wich conflict
              type => kind of error
              msgsum => short msg error
              msg => full error
             }
            }

=cut

sub report_error {
	my ($self, $u_rpm, $error) = @_;
	if (defined $error->{id} && exists $self->{urpm}{depslist}[$error->{id}]) {
		$error->{fullname} = $self->{urpm}{depslist}[$error->{id}]->fullname . " [".$self->{id_level}{$error->{id}}."]";
	}
	
	if (!defined $self->{error}{$u_rpm}) {
			my $pkg = $self->{urpm}{depslist}[$u_rpm];
			$self->{error}{$u_rpm}{fullname} = $pkg->fullname;
			if ($pkg->arch ne 'src') { 
				$self->{error}{$u_rpm}{sourcerpm} = $pkg->sourcerpm;
				($self->{error}{$u_rpm}{sname}) = $pkg->sourcerpm =~ /^(.*)-(.*)-(.*)\.src\.rpm$/;
			} else {
				$self->{error}{$u_rpm}{sname} = $pkg->name;	
			}
			$self->{error}{$u_rpm}{packager} = $self->pkg_lastpackager($pkg) || "unknonw";
			$self->{error}{$u_rpm}{level} = $self->{id_level}{$u_rpm};
	}
	push @{$self->{error}{$u_rpm}{err}}, { %{$error} };
}

=item * urpmchecker::explanation_error(@ERR);

Return an array of explanation for error index in @ERR.
Without @ERR, return know ERR.

=cut

sub explanation_error {
	my %err = (
DIRO =>
"A file is installed but the rpm which parent directory is in
repository level highter. To cleanly install, you should have this level.",
DIRM =>
"A file is installed by a rpm but no rpm own parent dir.
This mean uninstall the rpm, the parent dir will not be removed.",
CFL =>
"Those rpms contains same file, but files are differents, and those rpm do
not explicitly conflict. rpm will be unable to handle this safety.",
CFLT => 
"Those rpms contains same file, but files have differents type
(dir vs link, file vs link), those rpm do not explicitly conflict,
rpm will be unable to handle this safety.",
CFLP =>
"Those rpms contains same files but without same permissions.
If those files come from server apps, or contains secret password, 
it can create a security hole.",
CFLO =>
"Those rpms contains same files but files are not own by same user or group.
If those files come from server apps, or contains secret password,
it can create a security hole.",
DEP =>
"This requirement is missing in the distrib, installation or
building is impossible, maybe it need a rebuild.",
DEPO =>
"This requirement is present but in a sublevel repository, you can
install or build it only if you have the second repository in your config.",
DEPB =>
"The package depend on an old o an obsolete package, if thie requirement is remove,
the package will become uninstallable.",
OBS =>
"The package in level x is always obsolete by another in level y, 
then why it is in distro? If a package obsolete itself, there is a problem :)",
BIN =>
"This src package does produce any binary...",
BINO =>
"This src package produce binary but in another repository.",
SRC =>
"This binary does not have a src.",
SRCO =>
"This binary have a src but in another repository.",
VER =>
"An highter version of this package was found, so it shoulb be removed.",
VERM =>
"Two binary from same source was found, but do not come from same version
of source package, this mean the rpm is obsolete, or old. You should check
if this package should be remove.",
VERT =>
"This version of rpm is list twice, it will never update somethings.",
REB =>
"This src need to be build, binary are not found, or not up to date.",
	);
	@_ or return sort keys %err;
	map { "$_\n" . $err{$_}."\n" } @_
}

sub parse_rc {
	my ($file) = @_;
	my $rpmvalue;
     local *RC;
	open RC, $file or return 0;
     local $_;
	while (<RC>) {
		s/#.*//;         # no comments
          s/^\s+//;        # no leading white
          s/\s+$//;        # no trailing white
		length or next;
		my ($tag, $rest) = m/^([^:]*):(.*)$/;
		$tag =~ /^(arch_canon)$/ and do {
			my ($key, $compat, $val) = $rest =~ m/^\s*([^:]*):\s*(\S*)\s*(\S*)/;
			$rpmvalue->{$tag}{$key}{$compat} = $val || 1;
			next;
			};
		$tag =~ /^(arch_compat)|(buildarch_compat)$/ and do {
			my ($key, $rest) = $rest =~ m/^\s*([^:]*):\s*(.*)\s*$/;
			$rpmvalue->{$tag}{$key} = $_ foreach split(/\s+/, $rest);
			next;
			};
	}	
	close RC;
	$rpmvalue
}

sub rpmcompat {
	my ($self, $rpm_arch, $type, $os_arch) = @_;
	my $level = 1;
	$os_arch or $os_arch = $self->{arch};
	# this function is for binary, and we want the reference system !
	$os_arch or return 1;
	$rpm_arch =~ /src$/ and return 1;
	$rpm_arch eq $os_arch and return $level;
	while (exists $self->{rpmvalues}{$type}{$os_arch}) {
		$level++;
		$self->{rpmvalues}{$type}{$os_arch} eq $rpm_arch and return $level;
		$os_arch = $self->{rpmvalues}{$type}{$os_arch};
	}
	0
}

sub rpmisbuildable {
	my ($self, $pkg, $o_os_arch) = @_;
	$o_os_arch ||= $o_os_arch = $self->{arch} or return 1;
	my ($temp) = $self->rpmcompat($pkg->buildarchs); # FIXME: missing $rpm_arch, $typeo parameters


	$temp && $self->rpmcompat($pkg->buildarchs, 'buildarch_compat', $o_os_arch) == 0 and return 0;
	$pkg->excludearchs && any { /\Q$o_os_arch/ } $pkg->excludearchs and return 0;
	$pkg->exclusivearchs && !any { /\Q$o_os_arch/ } $pkg->exclusivearchs and return 0;
	1	
}

=item * pkg_is_conflict($pkg1,$pkg2)

Return 1 is thoses package conflict.
$pkg1 and $pkg2 can bee same package, it will check if it
don't conflict himself

=cut

sub pkg_is_conflict {
    my ($un, $deux) = @_;

    $un eq $deux and do {
        foreach my $c ($un->conflicts) {
            foreach ($deux->provides) {
                URPM::ranges_overlap($c, $_) and return 1;
            }
        }
        return 0;
    };
    
    foreach my $pkg ($un, $deux) {
        foreach my $c ($pkg->conflicts) {
            foreach ([ difference2([ $un, $deux ], [ $pkg ]) ]->[0]->provides) {
                URPM::ranges_overlap($c, $_) and return 1;
            }
        }
    }
    0
}

=item * $urpmchecker->pkg_need();

$state = {
    $rpm => rpm to look, # or
    @o_select => allready find,
    $level => recursion level,
    $rpmtop => id rpm init,
    $rpmfather => rpm which need,
    $dep => the dep name,
}

=cut

sub pkg_need {
    my ($self, $state, $callback) = @_;
    $state->{rpmtop} ||= $state->{rpm};
    $state->{o_select} ||= [];
    
    defined $callback and &$callback($state);
    
    push(@{$state->{o_select}},$state->{rpm});
    my @dep = $self->{urpm}{depslist}[$state->{rpm}]->requires;
    my $count=-1;
    foreach my $req (grep { $_ !~ /\|/  } @{$self->{deps}{$state->{rpm}}}) {
        $count++;
        $req =~ /^NOT/ and do {
             defined $callback and &$callback({ 
                    dep => $dep[$count],
                    rpmfather => $state->{rpm},
                    rpmtop => $state->{rpmtop},
                    o_select => $state->{o_select},
                    level => $state->{level} + 1,
                 });
            next;
        };
        any { $_ eq $req } @{$state->{o_select}} and do {
            defined $callback and &$callback({
                rpm => $req,
                o_select => $state->{o_select},
                rpmfather => $state->{rpm},
                rpmtop => $state->{rpmtop},
                level => $state->{level} + 1,
                dep => $dep[$count],                    
                });
            next;
        };
        $self->pkg_need({
                rpm => $req,
                o_select => $state->{o_select},
                rpmfather => $state->{rpm},
                rpmtop => $state->{rpmtop},
                level => $state->{level} + 1,
                dep => $dep[$count],
                makechoice => $state->{makechoice},
            }, $callback);
    }
    foreach my $req (grep { m/\|/ } @{$self->{deps}{$state->{rpm}}}) {
        my $choice;
        if (defined $state->{makechoice}) {
            $choice = $state->{makechoice}(split('\|', $req));
        } else {
            # See if one is allready in list:
            my @try = intersection(@{$state->{o_select}}, split('\|', $req));
            scalar(@try) or @try = split('\|', $req);
            my %select;
            foreach my $possible (@try) {
                foreach my $selected (@{$state->{o_select}}) {
                    any { $_ eq $possible } @{$self->{conflicts}{$selected}} or $select{$possible}=1;
                }
            }
            keys %select or $select{$_}=1 foreach (@try);
            $choice = $self->find_best_pkg(keys %select);
        }
        foreach my $choicefin (split('\|', $choice)) {
            any { $_ eq $choicefin } @{$state->{o_select}} and do {
                defined $callback and &$callback({
                    rpm => $choicefin,
                    o_select => $state->{o_select},
                    rpmfather => $state->{rpm},
                    rpmtop => $state->{rpmtop},
                    level => $state->{level} + 1,
                    dep => $dep[$count],
                    possible => [split('\|', $req)], 
                    });
                next;
            };
            $self->pkg_need({
                rpm => $choicefin,
                o_select => $state->{o_select},
                rpmfather => $state->{rpm},
                rpmtop => $state->{rpmtop},
                level => $state->{level} + 1,
                dep => $dep[$count],
                makechoice => $state->{makechoice},
                }, $callback);
        }
    }
    $state->{level} ? () : uniq(@{$state->{o_select}})
}

sub pkg_lastpackager {
	my ($_self, $pkg) = @_;
	my @c = $pkg->changelog_name;
	my $ch = $c[0];
	$ch =~ s/ \S$//;
     $ch =~ s/>.*/>/;
	$ch
}

sub idlevel {
    my ($self,$id) = @_;
    $self->{id_level}{$id}
}

sub idrep {
    my ($self,$id) = @_;
    $self->{id_rep}{$id}
}

sub idupdate {
    my ($self,$id) = @_;
    $self->{repository}[$self->idrep($id)]{update}
}

=item $self->is_best_pkg($id)

Return 0 if package is the best choice
Return 1 if a better package is found
Return 2 if it is obsoletes

=cut

sub is_best_pkg {
    my ($self,$id) = @_;
    any {
        $self->idupdate($_) > $self->idupdate($id) || ($self->idupdate($_) == $self->idupdate($id) && $self->idrep($_) <= $self->idrep($id)) and $_
        } @{$self->{obsoletes}{$id}} and return 2;
    $id != $self->find_best_pkg(@{$self->{rpm}{binary}{$self->{urpm}{depslist}[$id]->name}})
        and return 1;
    0
}

sub find_best_pkg {
    my ($self, @list) = @_;
    my $best;
    ($best,@list)=grep { !/^NOT/ } @list;
    my $level = $self->idlevel($best);
    foreach my $u_rpm (@list) {
        $u_rpm =~ /^NOT/ and next;
        
        if ($self->idupdate($best) != $self->idupdate($u_rpm) || $self->idlevel($u_rpm) == $level) {
            $self->{urpm}{depslist}[$u_rpm]->compare_pkg($self->{urpm}{depslist}[$best]) > 0 and do {
                $best=$u_rpm;
                $level=$self->idlevel($best);
                next;
            };
        } elsif($self->idlevel($u_rpm) < $level) {
            $best=$u_rpm;
            $level=$self->idlevel($best);
        }
    }
    $best
};

package urpmchecker::query;
our @ISA = ();

1
